import platform
from cffi import FFI
from pathlib import Path

image_intensities_root_path = Path(__file__).parent
image_intensities_sources_path = image_intensities_root_path / "sources"

image_intensities_definitions = """
    extern struct intensity_data {
        double nw;
        double ne;
        double sw;
        double se;
        int error;
    } intensity_data;

    struct intensity_data jpeg_intensities(const char *file_name);
    struct intensity_data png_intensities(const char *file_name);
"""


def prepare_variant():
    header_path = (image_intensities_sources_path / "definitions.h").resolve()
    extra_kwargs = {}

    ffi = FFI()
    ffi.cdef("""
        typedef struct intensity_data {
            double nw;
            double ne;
            double sw;
            double se;
            int error;
        } intensity_data;

        struct intensity_data jpeg_intensities(const char *file_name);
        struct intensity_data png_intensities(const char *file_name);
    """)

    sources = [
        "sources/intensities.c",
        "sources/jpeg.c",
        "sources/png.c",

        "sources/turbojpeg/jsimd_none.c",
        "sources/turbojpeg/jchuff.c",
        "sources/turbojpeg/jcapimin.c",
        "sources/turbojpeg/jcapistd.c",
        "sources/turbojpeg/jccolor.c",
        "sources/turbojpeg/jcicc.c",
        "sources/turbojpeg/jccoefct.c",
        "sources/turbojpeg/jcinit.c",
        "sources/turbojpeg/jcdctmgr.c",
        "sources/turbojpeg/jcmainct.c",
        "sources/turbojpeg/jcmarker.c",
        "sources/turbojpeg/jcmaster.c",
        "sources/turbojpeg/jcomapi.c",
        "sources/turbojpeg/jcparam.c",
        "sources/turbojpeg/jcphuff.c",
        "sources/turbojpeg/jcprepct.c",
        "sources/turbojpeg/jcsample.c",
        "sources/turbojpeg/jctrans.c",
        "sources/turbojpeg/jdapimin.c",
        "sources/turbojpeg/jdapistd.c",
        "sources/turbojpeg/jdatadst.c",
        "sources/turbojpeg/jdatasrc.c",
        "sources/turbojpeg/jdcoefct.c",
        "sources/turbojpeg/jdcolor.c",
        "sources/turbojpeg/jddctmgr.c",
        "sources/turbojpeg/jdhuff.c",
        "sources/turbojpeg/jdicc.c",
        "sources/turbojpeg/jdinput.c",
        "sources/turbojpeg/jdmainct.c",
        "sources/turbojpeg/jdmarker.c",
        "sources/turbojpeg/jdmaster.c",
        "sources/turbojpeg/jdmerge.c",
        "sources/turbojpeg/jdphuff.c",
        "sources/turbojpeg/jdpostct.c",
        "sources/turbojpeg/jdsample.c",
        "sources/turbojpeg/jdtrans.c",
        "sources/turbojpeg/jerror.c",
        "sources/turbojpeg/jfdctflt.c",
        "sources/turbojpeg/jfdctfst.c",
        "sources/turbojpeg/jfdctint.c",
        "sources/turbojpeg/jidctflt.c",
        "sources/turbojpeg/jidctfst.c",
        "sources/turbojpeg/jidctint.c",
        "sources/turbojpeg/jidctred.c",
        "sources/turbojpeg/jquant1.c",
        "sources/turbojpeg/jquant2.c",
        "sources/turbojpeg/jutils.c",
        "sources/turbojpeg/jmemmgr.c",
        "sources/turbojpeg/jmemnobs.c",
        "sources/turbojpeg/jaricom.c",
        "sources/turbojpeg/jdarith.c",
        "sources/turbojpeg/jcarith.c",
        "sources/turbojpeg/turbojpeg.c",
        "sources/turbojpeg/transupp.c",
        "sources/turbojpeg/jdatadst-tj.c",
        "sources/turbojpeg/jdatasrc-tj.c",
        "sources/turbojpeg/rdbmp.c",
        "sources/turbojpeg/rdppm.c",
        "sources/turbojpeg/wrbmp.c",
        "sources/turbojpeg/wrppm.c",
    ]

    include_dirs = [  # -I
        'sources/',
        'sources/turbojpeg/',
    ]
    libraries = [  # -L
        'png',
    ]
    extra_compile_args = [
        "-std=c99",
        "-fPIC",
        "-O3",
        "-DPPM_SUPPORTED",
        "-DBMP_SUPPORTED",
    ]

    if "Windows" in platform.system():
        extra_kwargs["extra_link_args"] = ["/NODEFAULTLIB:MSVCRTD"]
        extra_kwargs["libraries"] = ["advapi32"]
    # end def

    ffi.set_source(
        'image_intensities._intensities',
        f'#include "{ str(header_path) }"',
        sources=sources,
        include_dirs=include_dirs,  # -I
        libraries=libraries,  # -L
        extra_compile_args=extra_compile_args,
        **extra_kwargs,
    )

    return ffi
# end def


intensities_ffi = prepare_variant()

if __name__ == "__main__":
    intensities_ffi.compile(verbose=True)



import image_intensities.classes as classes
import image_intensities.intensities as intensities
from .classes import Intensities
from .version import VERSION, __version__

__all__ = [
    'classes', 'intensities',
    'jpg_intensities', 'png_intensities', 'image_intensities',
    '__version__', 'VERSION'
]

try:
    from .intensities import jpg_intensities, png_intensities, image_intensities
except ImportError:
    from .pure_python import jpg_intensities, png_intensities, image_intensities
# end try



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# from luckydonaldUtils.logger import logging
from luckydonaldUtils.encoding import to_binary as b

__author__ = 'luckydonald'

# logger = logging.getLogger(__name__)
# if __name__ == '__main__':
#     logging.add_colored_handler(level=logging.DEBUG)
# # end if

# noinspection PyUnresolvedReferences
from ._intensities import ffi as __ffi, lib as __lib
from .classes import Intensities
from mimetypes import guess_type


class ErrorCode(Exception):
    error: int

    def __init__(self, *args, error=None, **kwargs):
        assert error is not None
        self.error = error
        super().__init__(*args, **kwargs)
    # end def
# end class


def _convert_struct_to_luma(struct) -> Intensities:
    """
    :raises ErrorCode: The library had an error.
    """
    if struct.error:
        raise ErrorCode(error=struct.error)
    # end if
    return Intensities(nw=struct.nw, ne=struct.ne, sw=struct.sw, se=struct.se)
# end def


def jpg_intensities(filename) -> Intensities:
    """
    :raises ErrorCode: The library had an error.
    """
    filename = b(filename)
    result_struct = __lib.jpeg_intensities(filename)
    return _convert_struct_to_luma(result_struct)
# end def


def png_intensities(filename) -> Intensities:
    """
    :raises ErrorCode: The library had an error.
    """
    filename = b(filename)
    result_struct = __lib.png_intensities(filename)
    return _convert_struct_to_luma(result_struct)
# end def


def image_intensities(filename: str) -> Intensities:
    """
    :raises ErrorCode: The library had an error.
    :raises NotImplementedError: The file is not png/jpg.
    """
    (mime_type, encoding) = guess_type(filename)
    if mime_type == 'image/png':
        return png_intensities(filename)
    elif mime_type == 'image/jpeg':
        return jpg_intensities(filename)
    else:
        raise NotImplementedError('Unknown mime.')
    # end if
# end def



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from pathlib import Path

with open(str(Path(__file__).parent.absolute() / 'VERSION.txt'), 'r') as f:
    version = f.read().strip()
# end with

__version__ = version
VERSION = version

if __name__ == '__main__':
    print(__version__)
# end if



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Union

from luckydonaldUtils.logger import logging

__author__ = 'luckydonald'

logger = logging.getLogger(__name__)
if __name__ == '__main__':
    logging.add_colored_handler(level=logging.DEBUG)
# end if


DEFAULT_DISTANCE = 0.25


class RGB(object):
    r: int
    g: int
    b: int

    def __init__(
        self,
        r: int = 0,
        g: int = 0,
        b: int = 0,
    ):
        self.r = r
        self.g = g
        self.b = b
    # end def

    def __repr__(self):
        return f"{self.__class__.__name__}(r={self.r!r}, g={self.g!r}, b={self.b!r})"
    # end def
# end class


class QuadrantSums(object):
    nw: RGB
    ne: RGB
    sw: RGB
    se: RGB

    # noinspection PyShadowingNames
    def __init__(
        self,
        nw: Union[None, RGB] = None,
        ne: Union[None, RGB] = None,
        sw: Union[None, RGB] = None,
        se: Union[None, RGB] = None,
    ):
        if nw is None:
            nw = RGB()
        # end if
        self.nw = nw
        if ne is None:
            ne = RGB()
        # end if
        self.ne = ne
        if sw is None:
            sw = RGB()
        # end if
        self.sw = sw
        if se is None:
            se = RGB()
        # end if
        self.se = se
    # end def

    def __repr__(self):
        return f"{self.__class__.__name__}(nw={self.nw!r}, ne={self.ne!r}, sw={self.sw!r}, se={self.se!r})"
    # end def
# end class


class Intensities(object):
    nw: float
    ne: float
    sw: float
    se: float

    def __init__(self, nw: float, ne: float, sw: float, se: float):
        self.nw = nw
        self.ne = ne
        self.sw = sw
        self.se = se
    # end def

    def equals(self, other):
        return (
            self.nw == other.nw and
            self.ne == other.ne and
            self.sw == other.sw and
            self.se == other.se
        )
    # end def

    __eq__ = equals

    def compare(self, other: 'Intensities', *, distance: float = DEFAULT_DISTANCE) -> bool:
        """
        Match by distance

        :param other: the Luma value to compare with.
        :param distance: suggested values: between 0.2 and 0.5
        :return:
        """
        return (
            self.nw - distance < other.nw < self.nw + distance and
            self.ne - distance < other.ne < self.ne + distance and
            self.sw - distance < other.sw < self.sw + distance and
            self.se - distance < other.se < self.se + distance
        )
    # end def

    def __repr__(self):
        return f"{self.__class__.__name__}(nw={self.nw!r}, ne={self.ne!r}, sw={self.sw!r}, se={self.se!r})"
    # end def
# end class



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
convert the image into the yuv colorspace, drop the u and v components,
and then average the y component over 4 evenly-spaced rectangles on the
image dimensions that gives you a floating point number
"""

from PIL import Image
from luckydonaldUtils.logger import logging

__author__ = 'luckydonald'

from .classes import Intensities, RGB, QuadrantSums

__all__ = ['classes', 'intensities', 'jpg_intensities', 'png_intensities', 'image_intensities']

logger = logging.getLogger(__name__)
if __name__ == '__main__':
    logging.add_colored_handler(level=logging.DEBUG)
# end if


def rgb_sums(img: Image.Image, log_prefix="") -> QuadrantSums:
    width, height = img.size
    pixels = list(img.getdata())
    sums = QuadrantSums()
    log_counter = -1  # more efficient then i % 10 == 0
    log_interval = height // 10
    for i in range(height):
        log_counter += 1
        if log_counter == log_interval:
            logger.debug(f'{log_prefix}Processing line {i} of {width}x{height} px image.')
            log_counter = 0
        # end if
        for j in range(width):
            nw = (i <= height / 2) and (j <= width / 2)
            # noinspection PyShadowingNames
            ne = (i <= height / 2) and (j >= width / 2)
            # noinspection PyShadowingNames
            sw = (i >= height / 2) and (j <= width / 2)
            # noinspection PyShadowingNames
            se = (i >= height / 2) and (j >= width / 2)
            img.load()
            pixel = pixels[i * width + j]
            r, g, b = pixel[0], pixel[1], pixel[2]

            if nw:
                sums.nw.r += r
                sums.nw.g += g
                sums.nw.b += b
            # end if

            if ne:
                sums.ne.r += r
                sums.ne.g += g
                sums.ne.b += b
            # end if

            if sw:
                sums.sw.r += r
                sums.sw.g += g
                sums.sw.b += b
            # end if

            if se:
                sums.se.r += r
                sums.se.g += g
                sums.se.b += b
            # end if
        # end for
    # end for
    return sums
# end def


# noinspection PyShadowingBuiltins
def _calculate_luma(dim: int, sum: RGB):
    return (
       (sum.r / dim * 0.2126) +
       (sum.g / dim * 0.7152) +
       (sum.b / dim * 0.0772)
    ) / 3.0
# end def


def sums_to_luma(sums: QuadrantSums, img: Image.Image) -> Luma:
    width, height = img.size

    dim = max(width * height / 4.0, 1)

    return Intensities(
        nw=_calculate_luma(dim, sums.nw),
        ne=_calculate_luma(dim, sums.ne),
        sw=_calculate_luma(dim, sums.sw),
        se=_calculate_luma(dim, sums.se),
    )
# end def


def image_intensities(filename: str) -> Intensities:
    from PIL.Image import open
    img = open(filename)
    result = rgb_sums(img)
    return sums_to_luma(result, img)
# end def


jpg_intensities = image_intensities
png_intensities = image_intensities



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from pathlib import Path
from setuptools import setup, find_packages


__author__ = 'luckydonald'

here = Path(__file__).parent.absolute()

with open(str(here / 'README.md')) as f:
    long_description = f.read()
# end with

with open(str(here / 'image_intensities/VERSION.txt')) as f:
    version = f.read().strip()
# end with

setup(
    name="image_intensities", version=version,
    description='Calculate image intensities, a database friendly alternative to image hashing..',
    long_description=long_description,
    long_description_content_type='text/markdown',
    cffi_modules=["build.py:intensities_ffi"],
    # The project's main homepage.
    url='https://github.com/luckydonald/image_intensities',
    project_urls={
        "Code": "https://github.com/luckydonald/image_intensities",
        "Issue tracker": "https://github.com/luckydonald/image_intensities/issues"
    },
    # Author details
    author='luckydonald',
    author_email='image_intensities+code@luckydonald.de',
    # Choose your license
    license='GPLv3+',
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',  # 2 - Pre-Alpha, 3 - Alpha, 4 - Beta, 5 - Production/Stable
        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Telecommunications Industry',
        'Topic :: Utilities',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Environment :: Plugins',
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        #'Programming Language :: Python :: 2',
        # 'Programming Language :: Python :: 2.6',
        #'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        # 'Programming Language :: Python :: 3.2',
        # 'Programming Language :: Python :: 3.3',
        # 'Programming Language :: Python :: 3.4',
        # 'Programming Language :: Python :: 3.5',
        # 'Programming Language :: Python :: 3.6',
        # 'Programming Language :: Python :: 3.7',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Unix',
    ],
    # What does your project relate to?
    keywords='image hash image hashing image intensities nw sw se ne derpibooru',
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=["tests"]),
    # packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    # List run-time dependencies here. These will be installed by pip when your
    # project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "cffi>=1.4.0",
        "luckydonald-utils>=0.73",  # general utils
        # "pytgbot>=4.1.1",  # telegram communication
    ],
    # https://cffi.readthedocs.io/en/latest/cdef.html
    setup_requires=["cffi>=1.4.0"],
    # List additional groups of dependencies here (e.g. development dependencies).
    # You can install these using the following syntax, for example:
    # $ pip install -e .[dev,test]
    # extras_require = {
    # 'dev': ['check-manifest'],
    # 'test': ['coverage'],
    # },
    # If there are data files included in your packages that need to be
    # installed, specify them here. If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    # package_data={
    # 'sample': ['package_data.dat'],
    # },
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages.
    # see http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data'    , ['data/data_file'])],
    include_package_data=True,
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
)
